struct node
{
	vector<node*> son;
	node* fail;
	vector<int> ends;
	int visits;
	
	node()
	{
		son = vector<node*> (26, (node*) NULL);
		fail = NULL;
		visits = 0;
	}
	
	node* get_son(char c)
	{
		if(son[c -'a'] == NULL) son[c - 'a'] = new node;
		return son[c-'a'];
	}
	
	~node()
	{
		for(int i=0; i<26; i++)
		{
			if(son[i] != NULL) delete son[i];
		}
	}
};


vector<int> aho_corasick(vector<string> pattern, string text)
{
	//Constructing the trie
	node *root = new node;
	for(int i=0; i<pattern.size(); i++)
	{
		node *cur = root;
		for(int j=0; j<pattern[i].size(); j++)
		{
			cur = cur->get_son(pattern[i][j]);
		}
		cur->ends.push_back(i);
	}
	
	//Constructing faillinks
	node *preroot = new node;
	for(int i=0; i<26; i++) preroot->son[i] = root;
	root->fail = preroot;
	queue<node*> bfs_queue;
	vector<node*> nodes;
	bfs_queue.push(root);
	nodes.push_back(root);
	while(!bfs_queue.empty())
	{
		node *cur = bfs_queue.front();
		bfs_queue.pop();
		for(int i=0; i<26; i++)
		{
			if(cur->son[i] == NULL) continue;
			node *back = cur->fail;
			while(back->son[i] == NULL) back = back->fail;
			cur->son[i]->fail = back->son[i];
			bfs_queue.push(cur->son[i]);
			nodes.push_back(cur->son[i]);
		}
	}
	
	//Searching the string
	node *cur = root;
	for(int i=0; i<text.size(); i++)
	{
		while(cur->son[text[i]-'a'] == NULL) cur = cur->fail;
		cur = cur->son[text[i]-'a'];
		cur->visits++;
	}
	
	//Extracting the result
	vector<int> result(pattern.size());
	for(int i=nodes.size()-1; i>=0; i--)
	{
		nodes[i]->fail->visits += nodes[i]->visits;
		for(int j=0; j<nodes[i]->ends.size(); j++)
		{
			result[nodes[i]->ends[j]] = nodes[i]->visits;
		}
	}
	
	//Cleaning up
	for(int i=1; i<26; i++) preroot->son[i] = NULL;
	delete preroot;
	
	return result;
}
